import numpy as np
import tensorflow as tf
from keras import Input, Model
from keras import backend as K

from spektral.layers import GlobalSumPool, GlobalAttnSumPool, GlobalAttentionPool, GlobalAvgPool, GlobalMaxPool

sess = K.get_session()
batch_size = 32
N = 11
F = 7


def _check_output_and_model_output_shapes(true_shape, model_shape):
    assert len(true_shape) == len(model_shape)
    for i in range(len(true_shape)):
        if true_shape[i] == N:
            assert model_shape[i] in {N, None}
        elif true_shape[i] == batch_size:
            assert model_shape[i] in {batch_size, None}
        else:
            assert model_shape[i] == true_shape[i]


def _test_single_mode(layer, **kwargs):
    X = np.random.normal(size=(N, F))
    X_in = Input(shape=(F,))
    layer_instance = layer(**kwargs)
    output = layer_instance(X_in)
    model = Model(X_in, output)
    sess.run(tf.global_variables_initializer())
    output = sess.run(model.output, feed_dict={X_in: X})
    assert output.shape == (1, kwargs.get('channels', F))
    assert output.shape == layer_instance.compute_output_shape(X.shape)
    _check_output_and_model_output_shapes(output.shape, model.output_shape)


def _test_batch_mode(layer, **kwargs):
    X = np.random.normal(size=(batch_size, N, F))
    X_in = Input(shape=(N, F))
    layer_instance = layer(**kwargs)
    output = layer_instance(X_in)
    model = Model(X_in, output)
    sess.run(tf.global_variables_initializer())
    output = sess.run(model.output, feed_dict={X_in: X})
    assert output.shape == (batch_size, kwargs.get('channels', F))
    assert output.shape == layer_instance.compute_output_shape(X.shape)
    _check_output_and_model_output_shapes(output.shape, model.output_shape)


def _test_graph_mode(layer, **kwargs):
    X = np.random.normal(size=(batch_size * N, F))
    S = np.repeat(np.arange(batch_size), N).astype(np.int)
    X_in = Input(shape=(F,))
    S_in = Input(shape=(), dtype=S.dtype)
    layer_instance = layer(**kwargs)
    output = layer_instance([X_in, S_in])
    model = Model([X_in, S_in], output)
    sess.run(tf.global_variables_initializer())
    output = sess.run(model.output, feed_dict={X_in: X, S_in: S})
    assert output.shape == (batch_size, kwargs.get('channels', F))
    # When creating actual graph, the bacth dimension is None
    assert output.shape[1:] == layer_instance.compute_output_shape([X.shape, S.shape])[1:]
    _check_output_and_model_output_shapes(output.shape, model.output_shape)


def test_global_sum_pool():
    _test_single_mode(GlobalSumPool)
    _test_batch_mode(GlobalSumPool)
    _test_graph_mode(GlobalSumPool)


def test_global_avg_pool():
    _test_single_mode(GlobalAvgPool)
    _test_batch_mode(GlobalAvgPool)
    _test_graph_mode(GlobalAvgPool)


def test_global_max_pool():
    _test_single_mode(GlobalMaxPool)
    _test_batch_mode(GlobalMaxPool)
    _test_graph_mode(GlobalMaxPool)


def test_global_node_attention_pool():
    _test_single_mode(GlobalAttnSumPool)
    _test_batch_mode(GlobalAttnSumPool)
    _test_graph_mode(GlobalAttnSumPool)


def test_global_attention_pool():
    F_ = 10
    assert F_ != F
    _test_single_mode(GlobalAttentionPool, channels=F_)
    _test_batch_mode(GlobalAttentionPool, channels=F_)
    _test_graph_mode(GlobalAttentionPool, channels=F_)



from spektral.layers import GraphConv, ChebConv, EdgeConditionedConv, GraphAttention, GraphConvSkip, ARMAConv, APPNP, \
    GraphSageConv, GINConv
from keras import backend as K, Model, Input
import numpy as np
import tensorflow as tf

SINGLE, BATCH, MIXED = 1, 2, 3  # Single, batch, mixed
LAYER_K_, MODES_K_, KWARGS_K_ = 'layer', 'modes', 'kwargs'
TESTS = [
    {
        LAYER_K_: GraphConv,
        MODES_K_: [SINGLE, BATCH, MIXED],
        KWARGS_K_: {'channels': 8, 'activation': 'relu'}
    },
    {
        LAYER_K_: ChebConv,
        MODES_K_: [SINGLE, BATCH, MIXED],
        KWARGS_K_: {'channels': 8, 'activation': 'relu'}
    },
    {
        LAYER_K_: GraphSageConv,
        MODES_K_: [SINGLE],
        KWARGS_K_: {'channels': 8, 'activation': 'relu'}
    },
    {
        LAYER_K_: EdgeConditionedConv,
        MODES_K_: [BATCH],
        KWARGS_K_: {'channels': 8, 'activation': 'relu', 'edges': True}
    },
    {
        LAYER_K_: GraphAttention,
        MODES_K_: [SINGLE, BATCH, MIXED],
        KWARGS_K_: {'channels': 8, 'activation': 'relu'}
    },
    {
        LAYER_K_: GraphConvSkip,
        MODES_K_: [SINGLE, BATCH, MIXED],
        KWARGS_K_: {'channels': 8, 'activation': 'relu'}
    },
    {
        LAYER_K_: ARMAConv,
        MODES_K_: [SINGLE, BATCH, MIXED],
        KWARGS_K_: {'channels': 8, 'activation': 'relu', 'T': 2, 'K': 2, 'recurrent': True}
    },
    {
        LAYER_K_: APPNP,
        MODES_K_: [SINGLE, BATCH, MIXED],
        KWARGS_K_: {'channels': 8, 'activation': 'relu', 'mlp_channels': 16, 'H': 2}
    },
    {
        LAYER_K_: GINConv,
        MODES_K_: [SINGLE],
        KWARGS_K_: {'channels': 8, 'activation': 'relu', 'n_hidden_layers': 1}
    }
]

sess = K.get_session()
batch_size = 32
N = 11
F = 7
S = 3

A = np.ones((N, N))
X = np.random.normal(size=(N, F))
E = np.random.normal(size=(N, N, S))


def _test_single_mode(layer, **kwargs):
    A_in = Input(shape=(None,))
    X_in = Input(shape=(F,))

    layer_instance = layer(**kwargs)
    output = layer_instance([X_in, A_in])
    model = Model([X_in, A_in], output)

    sess.run(tf.global_variables_initializer())
    output = sess.run(model.output, feed_dict={X_in: X, A_in: A})

    assert output.shape == (N, kwargs['channels'])


def _test_batch_mode(layer, **kwargs):
    A_batch = np.stack([A] * batch_size)
    X_batch = np.stack([X] * batch_size)

    A_in = Input(shape=(N, N))
    X_in = Input(shape=(N, F))
    inputs = [X_in, A_in]
    feed_dict = {X_in: X_batch, A_in: A_batch}

    if kwargs.get('edges'):
        kwargs.pop('edges')
        E_batch = np.stack([E] * batch_size)
        E_in = Input(shape=(N, N, S))
        inputs.append(E_in)
        feed_dict[E_in] = E_batch

    layer_instance = layer(**kwargs)
    output = layer_instance(inputs)
    model = Model(inputs, output)

    sess.run(tf.global_variables_initializer())
    output = sess.run(model.output, feed_dict=feed_dict)

    assert output.shape == (batch_size, N, kwargs['channels'])


def _test_mixed_mode(layer, **kwargs):
    X_batch = np.stack([X] * batch_size)

    A_in = Input(shape=(N,))
    X_in = Input(shape=(N, F))
    inputs = [X_in, A_in]
    feed_dict = {X_in: X_batch, A_in: A}

    layer_instance = layer(**kwargs)
    output = layer_instance(inputs)
    model = Model(inputs, output)

    sess.run(tf.global_variables_initializer())
    output = sess.run(model.output, feed_dict=feed_dict)

    assert output.shape == (batch_size, N, kwargs['channels'])


def _test_get_config(layer, **kwargs):
    if kwargs.get('edges'):
        kwargs.pop('edges')
    layer_instance = layer(**kwargs)
    config = layer_instance.get_config()
    assert layer(**config)


def test_layers():
    for test in TESTS:
        for mode in test[MODES_K_]:
            if mode == SINGLE:
                _test_single_mode(test[LAYER_K_], **test[KWARGS_K_])
            elif mode == BATCH:
                _test_batch_mode(test[LAYER_K_], **test[KWARGS_K_])
            elif mode == MIXED:
                _test_mixed_mode(test[LAYER_K_], **test[KWARGS_K_])
        _test_get_config(test[LAYER_K_], **test[KWARGS_K_])



from __future__ import absolute_import

import numpy as np
from keras import activations, initializers, regularizers, constraints
from keras import backend as K
from keras.constraints import Constraint
from keras.layers import Layer, Average, Concatenate


class Affinity(Layer):
    """
    Wrapper for affinity functions, used in graph matching.

    **Mode**: batch.

    **Input**

    - The input to this layer can be any combination of pairs of adjacency matrices,
    node attributes, and edge attributes as follows:
        - list of length 2, with source and target adjacency matrices.
        Shapes: `[(batch_size, num_nodes, num_nodes), (batch_size, num_nodes, num_nodes)]`;
        - list of length 4, with source and target adjacency matrices, and only
        one type of attributes (node or edge). Shapes: `[(batch_size, num_nodes, num_nodes),
        (batch_size, num_nodes, num_nodes), (batch_size, num_nodes, num_features),
        (batch_size, num_nodes, num_features)]`.
        In this case, specify which of the two types of attribute is being passed
        by setting the corresponding `*_features_dim` flag appropriately.
        - list of length 6, with source and target adjacency matrices and features.
        Shapes: `[(batch_size, num_nodes, num_nodes),
        (batch_size, num_nodes, num_nodes), (batch_size, num_nodes, num_node_features),
        (batch_size, num_nodes, num_node_features), (batch_size, num_nodes, num_edge_features),
        (batch_size, num_nodes, num_edge_features)]`.

    **Output**

    - rank 2 tensor of shape `(input_dim_1, input_dim_1)`

    :param affinity_function: a function computing affinity tensors between
    graphs. The function will be called as:
    ```
    affinity_function(adj_src, adj_target, nf_src, nf_target, ef_src, ef_target, N=num_nodes, F=node_features_dim, S=edge_features_dim)
    ```.
    :param num_nodes: number of nodes in the graphs. It will be passed as 
    keyword argument to the affinity function (with key `N`);
    :param node_features_dim: number of node attributes. It will be passed as
    keyword argument to the affinity function (with key `F`);
    :param edge_features_dim: number of edge attributes. It will be passed as
    keyword argument to the affinity function (with key `S`);
    :param kwargs: optional arguments for `Layer`.
    """
    def __init__(self,
                 affinity_function,
                 num_nodes=None,
                 node_features_dim=None,
                 edge_features_dim=None,
                 **kwargs):
        self.affinity_function = affinity_function
        self.num_nodes = num_nodes
        self.node_features_dim = node_features_dim
        self.edge_features_dim = edge_features_dim
        super(Affinity, self).__init__(**kwargs)

    def build(self, input_shape):
        self.built = True

    def call(self, inputs):
        adj_src, adj_target, nf_src, nf_target, ef_src, ef_target = (None,) * 6
        if len(inputs) == 2:
            adj_src, adj_target = inputs
        elif len(inputs) == 4:
            if self.node_features_dim is not None and self.edge_features_dim is None:
                adj_src, adj_target, nf_src, nf_target = inputs
            elif self.node_features_dim is None and self.edge_features_dim is not None:
                adj_src, adj_target, ef_src, ef_target = inputs
            else:
                raise ValueError('Only four input tensors were passed, but '
                                 'it was not possible to interpret their '
                                 'meaning. If you passed node features, set '
                                 'num_node_features accordingly and leave '
                                 'num_edge_features=None (and vice-versa for '
                                 'edge features).')
        elif len(inputs) == 6:
            adj_src, adj_target, nf_src, nf_target, ef_src, ef_target = inputs
        else:
            raise ValueError('An even number of input tensors between 2 and 6 '
                             'is expected, representing input and output '
                             'adjacency matrices (mandatory), node features, '
                             'and edge features.')
        return self.affinity_function(adj_src, adj_target,
                                      nf_src, nf_target,
                                      ef_src, ef_target,
                                      N=self.num_nodes,
                                      F=self.node_features_dim,
                                      S=self.edge_features_dim)

    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self, **kwargs):
        config = {}
        base_config = super(Affinity, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class InnerProduct(Layer):
    """
    Computes the inner product between elements of a given 2d tensor \(x\): 
    $$
        \\langle x, x \\rangle = xx^T.
    $$

    **Mode**: single.

    **Input**

    - rank 2 tensor of shape `(input_dim_1, input_dim_2)` (e.g. node features
    of shape `(num_nodes, num_features)`);

    **Output**

    - rank 2 tensor of shape `(input_dim_1, input_dim_1)`

    :param trainable_kernel: add a trainable square matrix between the inner
    product (i.e., `x.dot(w).dot(x.T)`);
    :param activation: activation function to use;
    :param kernel_initializer: initializer for the kernel matrix;
    :param kernel_regularizer: regularization applied to the kernel;
    :param activity_regularizer: regularization applied to the output;
    :param kernel_constraint: constraint applied to the kernel;
    """
    def __init__(self,
                 trainable_kernel=False,
                 activation=None,
                 kernel_initializer='glorot_uniform',
                 kernel_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 **kwargs):
        if 'input_shape' not in kwargs and 'input_dim' in kwargs:
            kwargs['input_shape'] = (kwargs.pop('input_dim'),)
        super(InnerProduct, self).__init__(**kwargs)
        self.trainable_kernel = trainable_kernel
        self.activation = activations.get(activation)
        self.kernel_initializer = initializers.get(kernel_initializer)
        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.activity_regularizer = regularizers.get(activity_regularizer)
        self.kernel_constraint = constraints.get(kernel_constraint)

    def build(self, input_shape):
        assert len(input_shape) >= 2
        if self.trainable_kernel:
            features_dim = input_shape[-1]
            self.kernel = self.add_weight(shape=(features_dim, features_dim),
                                          initializer=self.kernel_initializer,
                                          name='kernel',
                                          regularizer=self.kernel_regularizer,
                                          constraint=self.kernel_constraint)
        self.built = True

    def call(self, inputs):
        if self.trainable_kernel:
            output = K.dot(K.dot(inputs, self.kernel), K.transpose(inputs))
        else:
            output = K.dot(inputs, K.transpose(inputs))
        if self.activation is not None:
            output = self.activation(output)
        return output

    def compute_output_shape(self, input_shape):
        if len(input_shape) == 2:
            return (None, None)
        else:
            return input_shape[:-1] + (input_shape[-2], )

    def get_config(self, **kwargs):
        config = {}
        base_config = super(InnerProduct, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class MinkowskiProduct(Layer):
    """
    Computes the hyperbolic inner product between elements of a given 2d tensor
    \(x\): 
    $$
        \\langle x, x \\rangle = x \,
        \\begin{pmatrix}
            I_{d\\times d} & 0 \\\\ 0 & -1
        \\end{pmatrix} \\,x^T.
    $$

    **Mode**: single.

    **Input**

    - rank 2 tensor of shape `(input_dim_1, input_dim_2)` (e.g. node features
    of shape `(num_nodes, num_features)`);

    **Output**

    - rank 2 tensor of shape `(input_dim_1, input_dim_1)`

    :param input_dim_1: first dimension of the input tensor; set this if you
    encounter issues with shapes in your model, in order to provide an explicit
    output shape for your layer.
    :param activation: activation function to use;
    :param activity_regularizer: regularization applied to the output;
    """
    def __init__(self,
                 input_dim_1=None,
                 activation=None,
                 activity_regularizer=None,
                 **kwargs):
        if 'input_shape' not in kwargs and 'input_dim' in kwargs:
            kwargs['input_shape'] = (kwargs.pop('input_dim'),)
        super(MinkowskiProduct, self).__init__(**kwargs)
        self.input_dim_1 = input_dim_1
        self.activation = activations.get(activation)
        self.activity_regularizer = regularizers.get(activity_regularizer)

    def build(self, input_shape):
        assert len(input_shape) >= 2
        self.built = True

    def call(self, inputs):
        F = K.int_shape(inputs)[-1]
        minkowski_prod_mat = np.eye(F)
        minkowski_prod_mat[-1, -1] = -1.
        minkowski_prod_mat = K.constant(minkowski_prod_mat)
        output = K.dot(inputs, minkowski_prod_mat)
        output = K.dot(output, K.transpose(inputs))
        output = K.clip(output, -10e9, -1.)

        if self.activation is not None:
            output = self.activation(output)

        return output

    def compute_output_shape(self, input_shape):
        if len(input_shape) == 2:
            if self.input_dim_1 is None:
                return (None, None)
            else:
                return (self.input_dim_1, self.input_dim_1)
        else:
            return input_shape[:-1] + (input_shape[-2], )

    def get_config(self, **kwargs):
        config = {}
        base_config = super(MinkowskiProduct, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class CCMProjection(Layer):
    """
    Projects a tensor to a CCM depending on the value of `r`. Optionally, 
    `r` can be learned via backpropagation.

    **Input**

    - tensor of shape `(batch_size, input_dim)`.

    **Output**

    - tensor of shape `(batch_size, input_dim)`, where each sample along the
    0th axis is projected to the CCM.

    :param r: radius of the CCM. If r is a number, then use it as fixed
    radius. If `r='spherical'`, use a trainable weight as radius, with a
    positivity constraint. If `r='hyperbolic'`, use a trainable weight
    as radius, with a negativity constraint. If `r=None`, use a trainable
    weight as radius, with no constraints (points will be projected to the
    correct manifold based on the sign of the weight).
    :param kernel_initializer: initializer for the kernel matrix;
    :param kernel_regularizer: regularization applied to the kernel matrix;
    :param kernel_constraint: constraint applied to the kernel matrix.
    """
    def __init__(self,
                 r=None,
                 kernel_initializer='glorot_uniform',
                 kernel_regularizer=None,
                 kernel_constraint=None, **kwargs):
        super(CCMProjection, self).__init__(**kwargs)
        self.radius = r
        self.kernel_initializer = initializers.get(kernel_initializer)
        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        if self.radius == 'spherical':
            self.kernel_constraint = self.Pos()
        elif self.radius == 'hyperbolic':
            self.kernel_constraint = self.Neg()
        else:
            self.kernel_constraint = constraints.get(kernel_constraint)

    def build(self, input_shape):
        assert len(input_shape) >= 2
        if self.radius in {'spherical', 'hyperbolic'} or self.radius is None:
            self.radius = self.add_weight(shape=(),
                                          initializer=self.kernel_initializer,
                                          name='radius',
                                          regularizer=self.kernel_regularizer,
                                          constraint=self.kernel_constraint)
        else:
            self.radius = K.constant(self.radius)
        self.built = True

    def call(self, inputs):
        zero = K.constant(0.)

        # Spherical clip
        spherical_clip = self.radius * K.l2_normalize(inputs, -1)
        # Hyperbolic clip
        free_components = inputs[..., :-1]
        bound_component = K.sqrt(K.sum(free_components ** 2, -1)[..., None] + (self.radius ** 2))
        hyperbolic_clip = K.concatenate((free_components, bound_component), -1)

        lt_cond = K.less(self.radius, zero)
        lt_check = K.switch(lt_cond, hyperbolic_clip, inputs)

        gt_cond = K.greater(self.radius, zero)
        output = K.switch(gt_cond, spherical_clip, lt_check)

        return output

    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self, **kwargs):
        config = {}
        base_config = super(CCMProjection, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    class Pos(Constraint):
        """Constrains a scalar weight to be positive.
        """

        def __call__(self, w):
            return K.maximum(w, K.epsilon())

    class Neg(Constraint):
        """Constrains a scalar weight to be negative.
        """

        def __call__(self, w):
            return K.minimum(w, -K.epsilon())


class CCMMembership(Layer):
    """
    Computes the membership of the given points to a constant-curvature
    manifold of radius `r`, as: 
    $$
        \\mu(x) = \\mathrm{exp}\\left(\\cfrac{-\\big( \\langle \\vec x, \\vec x \\rangle - r^2 \\big)^2}{2\\sigma^2}\\right).
    $$

    If `r=0`, then \(\\mu(x) = 1\).
    If more than one radius is given, inputs are evenly split across the 
    last dimension and membership is computed for each radius-slice pair.
    The output membership is returned according to the `mode` option.

    **Input**

    - tensor of shape `(batch_size, input_dim)`;

    **Output**

    - tensor of shape `(batch_size, output_size)`, where `output_size` is
    computed according to the `mode` option;.

    :param r: int ot list, radia of the CCMs.
    :param mode: 'average' to return the average membership across CCMs, or
    'concat' to return the membership for each CCM concatenated;
    :param sigma: spread of the membership curve;
    """
    def __init__(self, r=1., mode='average', sigma=1., **kwargs):
        super(CCMMembership, self).__init__(**kwargs)
        if isinstance(r, int) or isinstance(r, float):
            self.r = [r]
        elif isinstance(r, list) or isinstance(r, tuple):
            self.r = r
        else:
            raise TypeError('r must be either a single value, or a list/tuple '
                            'of values.')
        possible_modes = {'average', 'concat'}
        if mode not in possible_modes:
            raise ValueError('Possible modes: {}'.format(possible_modes))
        self.mode = mode
        self.sigma = sigma

    def build(self, input_shape):
        self.built = True

    def call(self, inputs):
        output_part = []
        manifold_size = K.int_shape(inputs)[-1] // len(self.r)

        for idx, r_ in enumerate(self.r):
            start = idx * manifold_size
            stop = start + manifold_size
            part = inputs[..., start:stop]
            sign = np.sign(r_)
            if sign == 0.:
                # This is weird but necessary to make the layer differentiable
                output_pre = K.sum(inputs, -1, keepdims=True) * 0. + 1.
            else:
                free_components = part[..., :-1] ** 2
                bound_component = sign * part[..., -1:] ** 2
                all_components = K.concatenate((free_components, bound_component), -1)
                ext_product = K.sum(all_components, -1, keepdims=True)
                output_pre = K.exp(-(ext_product - sign * r_ ** 2) ** 2 / (2 * self.sigma ** 2))

            output_part.append(output_pre)

        if len(output_part) >= 2:
            if self.mode == 'average':
                output = Average()(output_part)
            elif self.mode == 'concat':
                output = Concatenate()(output_part)
            else:
                raise ValueError()  # Never gets here
        else:
            output = output_part[0]

        return output

    def compute_output_shape(self, input_shape):
        output_shape = input_shape[:-1] + (1, )
        return output_shape

    def get_config(self, **kwargs):
        config = {}
        base_config = super(CCMMembership, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class GaussianSampling(Layer):
    """
    A layer for sampling from a Gaussian distribution using the re-parametrization
    trick of Kingma et al., 2014.

    **Input**

    - tensor of shape `(batch_size, input_dim)` representing the means;
    - tensor of shape `(batch_size, input_dim)` representing the log-variances;

    **Output**

    - tensor of shape `(batch_size, input_dim)`, obtained by sampling from a
    Gaussian distribution with the given means and log-variances, using the
    re-parametrization trick;
        
    :param mean: mean of the Gaussian noise;
    :param std: standard deviation of the Gaussian noise.
    """
    def __init__(self, mean=0., std=1.0, **kwargs):
        super(GaussianSampling, self).__init__(**kwargs)
        self.mean = mean
        self.std = std

    def build(self, input_shape):
        assert len(input_shape) >= 2
        self.built = True

    def call(self, inputs):
        z_mean, z_log_var = inputs
        epsilon = K.random_normal(shape=K.shape(z_log_var),
                                  mean=self.mean,
                                  stddev=self.std)
        return z_mean + K.exp(0.5 * z_log_var) * epsilon

    def compute_output_shape(self, input_shape):
        return input_shape[0]

    def get_config(self, **kwargs):
        config = {}
        base_config = super(GaussianSampling, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class Bias(Layer):
    """
    A layer for adding a trainable bias vector (wrapper for `K.bias_add`).

    **Input**

    - tensor of shape `(batch_size, input_dim_1, ..., input_dim_n)`;

    **Output**

    - tensor of shape `(batch_size, input_dim_1, ..., input_dim_n)`;

    :param bias_initializer: initializer for the bias;
    :param bias_regularizer: regularizer for the bias;
    :param bias_constraint: constraint for the bias;

    """
    def __init__(self,
                 bias_initializer='zeros',
                 bias_regularizer=None,
                 bias_constraint=None,
                 **kwargs):
        super(Bias, self).__init__(**kwargs)
        self.bias_initializer = initializers.get(bias_initializer)
        self.bias_regularizer = regularizers.get(bias_regularizer)
        self.bias_constraint = constraints.get(bias_constraint)
        self.supports_masking = False

    def build(self, input_shape):
        assert len(input_shape) >= 2
        self.bias = self.add_weight(shape=input_shape[1:],
                                    initializer=self.bias_initializer,
                                    name='bias',
                                    regularizer=self.bias_regularizer,
                                    constraint=self.bias_constraint)
        self.built = True

    def call(self, inputs):
        return K.bias_add(inputs, self.bias)

    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self, **kwargs):
        config = {}
        base_config = super(Bias, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))




import numpy as np
import tensorflow as tf
from keras import Input, Model
from keras import backend as K

from spektral.layers import TopKPool, MinCutPool
import scipy.sparse as sp

sess = K.get_session()
batch_size = 3
N1, N2, N3 = 4, 5, 2
N = N1 + N2 + N3
F = 7


def _check_output_and_model_output_shapes(true_shape, model_shape):
    assert len(true_shape) == len(model_shape)
    for i in range(len(true_shape)):
        assert len(true_shape[i]) == len(model_shape[i])
        for j in range(len(true_shape[i])):
            assert model_shape[i][j] in {true_shape[i][j], None}


def _check_number_of_nodes(N_pool_expected, N_pool_true):
    if N_pool_expected is not None:
        assert N_pool_expected == N_pool_true or N_pool_true is None


def _test_single_mode(layer, **kwargs):
    A = np.ones((N, N))
    X = np.random.normal(size=(N, F))

    A_in = Input(shape=(None, ))
    X_in = Input(shape=(F,))

    layer_instance = layer(**kwargs)
    output = layer_instance([X_in, A_in])
    model = Model([X_in, A_in], output)
    sess.run(tf.global_variables_initializer())
    output = sess.run(model.output, feed_dict={X_in: X, A_in: A})
    X_pool, A_pool, mask = output

    if 'ratio' in kwargs.keys():
        N_exp = kwargs['ratio'] * N
    elif 'k' in kwargs.keys():
        N_exp = kwargs['k']
    else:
        raise ValueError('Need k or ratio.')
    N_pool_expected = np.ceil(N_exp)
    N_pool_true = A_pool.shape[-1]

    _check_number_of_nodes(N_pool_expected, N_pool_true)

    assert X_pool.shape == (N_pool_expected, F)
    assert A_pool.shape == (N_pool_expected, N_pool_expected)

    output_shape = [o.shape for o in output]
    _check_output_and_model_output_shapes(output_shape, model.output_shape)


def _test_graph_mode(layer, **kwargs):
    A = sp.block_diag([np.ones((N1, N1)), np.ones((N2, N2)), np.ones((N3, N3))]).todense()
    X = np.random.normal(size=(N, F))
    I = np.array([0] * N1 + [1] * N2 + [2] * N3).astype(int)

    A_in = Input(shape=(None, ))
    X_in = Input(shape=(F,))
    I_in = Input(shape=(), dtype=tf.int32)

    layer_instance = layer(**kwargs)
    output = layer_instance([X_in, A_in, I_in])
    model = Model([X_in, A_in, I_in], output)
    sess.run(tf.global_variables_initializer())
    output = sess.run(model.output, feed_dict={X_in: X, A_in: A, I_in: I})
    X_pool, A_pool, I_pool, mask = output

    N_pool_expected = np.ceil(kwargs['ratio'] * N1) + \
                      np.ceil(kwargs['ratio'] * N2) + \
                      np.ceil(kwargs['ratio'] * N3)
    N_pool_true = A_pool.shape[0]

    _check_number_of_nodes(N_pool_expected, N_pool_true)

    assert X_pool.shape == (N_pool_expected, F)
    assert A_pool.shape == (N_pool_expected, N_pool_expected)
    assert I_pool.shape == (N_pool_expected, )

    output_shape = [o.shape for o in output]
    _check_output_and_model_output_shapes(output_shape, model.output_shape)


def test_top_k_pool():
    _test_single_mode(TopKPool, ratio=0.5, return_mask=True)
    _test_graph_mode(TopKPool, ratio=0.5, return_mask=True)

def test_mincut_pool():
    _test_single_mode(MinCutPool, k=5, return_mask=True)





from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='spektral',
    version='0.0.12',
    packages=find_packages(),
    install_requires=['keras', 'networkx', 'pandas', 'joblib', 'matplotlib',
                      'tqdm', 'pygraphviz', 'numpy', 'scipy', 'requests',
                      'scikit-learn'],
    url='https://github.com/danielegrattarola/spektral',
    license='MIT',
    author='Daniele Grattarola',
    author_email='daniele.grattarola@gmail.com',
    description='A Python framework for relational representation learning',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3.5",
    ],
)



from __future__ import absolute_import

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
from matplotlib import animation

from .conversion import numpy_to_nx
from mpl_toolkits.mplot3d import Axes3D

HORIZONTAL_ROTATION = 30
VERTICAL_ROTATION = 5


# GRAPHS #######################################################################
def plot_nx(nx_graph, nf_name=None, ef_name=None, layout='spring_layout',
            labels=True, node_color='r', node_size=300, **kwds):
    """
    Plot the given Networkx graph.
    :param nx_graph: a Networkx graph
    :param nf_name: name of the relevant node feature to plot
    :param ef_name: name of the relevant edgee feature to plot
    :param layout: type of layout for networkx
    :param labels: plot labels
    :param node_color: color for the plotted nodes
    :param node_size: size of the plotted nodes
    :return: None
    """
    layout = deserialize_nx_layout(layout, nf_name=nf_name)
    pos = layout(nx_graph)
    nx.draw(nx_graph, pos, node_color=node_color, node_size=node_size, **kwds)
    if nf_name is not None:
        node_labels = nx.get_node_attributes(nx_graph, nf_name)
        if labels:
            nx.draw_networkx_labels(nx_graph, pos, labels=node_labels)
    if ef_name is not None:
        edge_labels = nx.get_edge_attributes(nx_graph, ef_name)
        if labels:
            nx.draw_networkx_edge_labels(nx_graph, pos, edge_labels=edge_labels)


def plot_numpy(adj, node_features=None, edge_features=None, nf_name=None,
               ef_name=None, layout='spring_layout', labels=True,
               node_color='r', node_size=300, ):
    """
    Converts a graph in matrix format (i.e. with adjacency matrix, node features
    matrix, and edge features matrix) to the Networkx format, then plots it with
    plot_nx().
    :param adj: np.array, adjacency matrix of the graph 
    :param node_features: np.array, node features matrix of the graph
    :param edge_features: np.array, edge features matrix of the graph
    :param nf_name: name to assign to the node features
    :param ef_name: name to assign to the edge features
    :param layout: type of layout for networkx
    :param labels: plot labels
    :param node_color: color for the plotted nodes
    :param node_size: size of the plotted nodes
    :return: None
    """
    if node_features is not None and nf_name is None:
        nf_name = 'nf'
    if edge_features is not None and ef_name is None:
        ef_name = 'ef'
    g = numpy_to_nx(adj, node_features, edge_features, nf_name, ef_name)
    plot_nx(g, nf_name, ef_name, node_color=node_color, node_size=node_size, layout=layout, labels=labels)


# 3D PLOTS   ###################################################################
def spherical_plot(data, r, keep=False, **kwargs):
    if data.shape[-1] != 3:
        raise ValueError('data must have shape (n_samples, 3).')
    u_s = np.linspace(0, 2 * np.pi, 100)
    v_s = np.linspace(0, np.pi, 100)
    x_s = r * np.outer(np.cos(u_s), np.sin(v_s))
    y_s = r * np.outer(np.sin(u_s), np.sin(v_s))
    z_s = r * np.outer(np.ones(np.size(u_s)), np.cos(v_s))
    if not keep:
        plt.gcf().add_subplot(111, projection='3d')
    ax = plt.gca()
    ax.set_ylim(-(r * 1.1), (r * 1.1))
    ax.set_xlim(-(r * 1.1), (r * 1.1))
    ax.set_zlim(-(r * 1.1), (r * 1.1))
    ax.plot_wireframe(x_s, y_s, z_s, rstride=10, cstride=10, color='k',
                      linewidth=0.5)
    ax.scatter(0, 0, 0, c='k')
    ax.scatter(data[:, 0], data[:, 1], data[:, 2], **kwargs)

    return ax


def hyperbolic_plot(data, r, **kwargs):
    if data.shape[-1] != 3:
        raise ValueError('data must have shape (n_samples, 3).')
    x_s = np.arange(-10, 10, 0.1)
    y_s = np.arange(-10, 10, 0.1)
    x_s, y_s = np.meshgrid(x_s, y_s)
    z_s = np.sqrt((x_s ** 2 + y_s ** 2) + r ** 2)
    plt.gcf().add_subplot(111, projection='3d')
    ax = plt.gca()
    ax.set_xlim(data[:, 0].min() - 1, data[:, 0].max() + 1)
    ax.set_ylim(data[:, 1].min() - 1, data[:, 1].max() + 1)
    ax.set_zlim(data[:, 2].min() - 1, data[:, 2].max() + 1)
    ax.plot_wireframe(x_s, y_s, z_s, rstride=10, cstride=10, color='k',
                      linewidth=0.5)
    ax.scatter(0, 0, 0, c='k')
    ax.scatter(data[:, 0], data[:, 1], data[:, 2], **kwargs)

    return ax


def euclidean_plot(data, **kwargs):
    if data.shape[-1] != 3:
        raise ValueError('data must have shape (n_samples, 3).')
    plt.gcf().add_subplot(111, projection='3d')
    ax = plt.gca()
    ax.set_xlim(data[:, 0].min() - 1, data[:, 0].max() + 1)
    ax.set_ylim(data[:, 1].min() - 1, data[:, 1].max() + 1)
    ax.set_zlim(data[:, 2].min() - 1, data[:, 2].max() + 1)
    ax.scatter(0, 0, 0, c='k')
    ax.view_init(15, 70)
    ax.scatter(data[:, 0], data[:, 1], data[:, 2], **kwargs)

    return ax


# ANIMATIONS ###################################################################
def histogram_animation(data, window_size=1, stride=1, frames=None,
                        filename=None, fps=30):
    def update_hist(frm):
        plt.cla()
        plt.xlim(data[:, ...].min(), data[:, ...].max())
        plt.hist(data[frm * stride: frm * stride + window_size, ...],
                 bins=100, color=plt.cm.winter(float(frm) / frames),
                 density=True)

    fig = plt.figure()
    plt.xlim(data[:, ...].min(), data[:, ...].max())
    plt.hist(data[:window_size, ...], bins=100, color=plt.cm.winter(0.),
             density=True)

    ani = animation.FuncAnimation(fig, update_hist, frames=frames)
    if filename is not None:
        ani.save(filename, writer='imagemagick', fps=fps)
    return ani


def spherical_animation(data, r, window_size=1, stride=1, frames=None,
                        filename=None, fps=30):
    if data.shape[-1] != 3:
        raise ValueError('data must have shape (n_samples, 3).')
    u_s = np.linspace(0, 2 * np.pi, 100)
    v_s = np.linspace(0, np.pi, 100)
    x_s = r * np.outer(np.cos(u_s), np.sin(v_s))
    y_s = r * np.outer(np.sin(u_s), np.sin(v_s))
    z_s = r * np.outer(np.ones(np.size(u_s)), np.cos(v_s))

    # Animation of embeddings as change happens
    def update_hist(frm):
        progress = float(frm) / frames
        start = frm * stride
        stop = start + window_size
        plt.cla()
        ax = plt.gca()
        ax.set_ylim(-(r * 1.1), (r * 1.1))
        ax.set_xlim(-(r * 1.1), (r * 1.1))
        ax.set_zlim(-(r * 1.1), (r * 1.1))
        ax.plot_wireframe(x_s, y_s, z_s, rstride=10, cstride=10, color='k',
                          linewidth=0.5)
        ax.scatter(0, 0, 0, c='k')
        ax.view_init(15, 70)
        ax.scatter(data[start:stop, 0], data[start:stop, 1],
                   data[start:stop, 2],
                   c=plt.cm.winter(progress),
                   marker='.', alpha=0.9)
        ax.view_init(15 + VERTICAL_ROTATION * progress,
                     70 + HORIZONTAL_ROTATION * progress)

    fig = plt.figure()
    plt.gcf().add_subplot(111, projection='3d')
    ax = plt.gca()
    ax.set_ylim(-(r * 1.1), (r * 1.1))
    ax.set_xlim(-(r * 1.1), (r * 1.1))
    ax.set_zlim(-(r * 1.1), (r * 1.1))
    ax.plot_wireframe(x_s, y_s, z_s, rstride=10, cstride=10, color='k',
                      linewidth=0.5)
    ax.scatter(0, 0, 0, c='k')
    ax.view_init(15, 70)
    ax.scatter(data[:window_size, 0],
               data[:window_size, 1],
               data[:window_size, 2],
               c=plt.cm.winter(0.), marker='.', alpha=0.9)

    ani = animation.FuncAnimation(fig, update_hist, frames=frames)
    if filename is not None:
        ani.save(filename, writer='imagemagick', fps=fps)

    return ani


def hyperbolic_animation(data, r, window_size=1, stride=1, frames=None,
                         filename=None, fps=30):
    if data.shape[-1] != 3:
        raise ValueError('data must have shape (n_samples, 3).')
    x_s = np.arange(-10, 10, 0.1)
    y_s = np.arange(-10, 10, 0.1)
    x_s, y_s = np.meshgrid(x_s, y_s)
    z_s = np.sqrt((x_s ** 2 + y_s ** 2) + r ** 2)

    # Animation of embeddings as change happens
    def update_hist(frm):
        progress = float(frm) / frames
        start = frm * stride
        stop = start + window_size
        plt.cla()
        ax = plt.gca()
        ax.set_xlim(data[:, 0].min() - 1,
                    data[:, 0].max() + 1)
        ax.set_ylim(data[:, 1].min() - 1,
                    data[:, 1].max() + 1)
        ax.set_zlim(data[:, 2].min() - 1,
                    data[:, 2].max() + 1)
        ax.plot_wireframe(x_s, y_s, z_s, rstride=10, cstride=10, color='k',
                          linewidth=0.5)
        ax.scatter(0, 0, 0, c='k')
        ax.view_init(15, 70)
        ax.scatter(data[start:stop, 0],
                   data[start:stop, 1],
                   data[start:stop, 2],
                   c=plt.cm.winter(progress),
                   marker='.', alpha=0.9)
        ax.view_init(15 + VERTICAL_ROTATION * progress,
                     70 + HORIZONTAL_ROTATION * progress)

    fig = plt.figure()
    plt.gcf().add_subplot(111, projection='3d')
    ax = plt.gca()
    ax.set_xlim(data[:, 0].min() - 1,
                data[:, 0].max() + 1)
    ax.set_ylim(data[:, 1].min() - 1,
                data[:, 1].max() + 1)
    ax.set_zlim(data[:, 2].min() - 1,
                data[:, 2].max() + 1)
    ax.plot_wireframe(x_s, y_s, z_s, rstride=10, cstride=10, color='k',
                      linewidth=0.5)
    ax.scatter(0, 0, 0, c='k')
    ax.view_init(15, 70)
    ax.scatter(data[:window_size, 0],
               data[:window_size, 1],
               data[:window_size, 2],
               c=plt.cm.winter(0.), marker='.', alpha=0.9)

    ani = animation.FuncAnimation(fig, update_hist, frames=frames)
    if filename is not None:
        ani.save(filename, writer='imagemagick', fps=fps)

    return ani


def euclidean_animation(data, window_size=1, stride=1, frames=None,
                        filename=None, fps=30):
    if data.shape[-1] != 3:
        raise ValueError('data must have shape (n_samples, 3).')

    # Animation of embeddings as change happens
    def update_hist(frm):
        progress = float(frm) / frames
        start = frm * stride
        stop = start + window_size
        plt.cla()
        ax = plt.gca()
        ax.set_xlim(data[:, 0].min() - 1,
                    data[:, 0].max() + 1)
        ax.set_ylim(data[:, 1].min() - 1,
                    data[:, 1].max() + 1)
        ax.set_zlim(data[:, 2].min() - 1,
                    data[:, 2].max() + 1)
        ax.scatter(0, 0, 0, c='k')
        ax.view_init(15, 70)
        ax.scatter(data[start:stop, 0],
                   data[start:stop, 1],
                   data[start:stop, 2],
                   c=plt.cm.winter(progress),
                   marker='.', alpha=0.9)
        ax.view_init(15 + VERTICAL_ROTATION * progress,
                     70 + HORIZONTAL_ROTATION * progress)

    fig = plt.figure()
    plt.gcf().add_subplot(111, projection='3d')
    ax = plt.gca()
    ax.set_xlim(data[:, 0].min() - 1,
                data[:, 0].max() + 1)
    ax.set_ylim(data[:, 1].min() - 1,
                data[:, 1].max() + 1)
    ax.set_zlim(data[:, 2].min() - 1,
                data[:, 2].max() + 1)
    ax.scatter(0, 0, 0, c='k')
    ax.view_init(15, 70)
    ax.scatter(data[:window_size, 0],
               data[:window_size, 1],
               data[:window_size, 2],
               c=plt.cm.winter(0.), marker='.', alpha=0.9)

    ani = animation.FuncAnimation(fig, update_hist, frames=frames)
    if filename is not None:
        ani.save(filename, writer='imagemagick', fps=fps)

    return ani


# Utils
def coordinates_layout_closure(nf_name):
    def coordinates_layout(nx_graph):
        return nx.get_node_attributes(nx_graph, nf_name)

    return coordinates_layout


def deserialize_nx_layout(layout, nf_name=None):
    if isinstance(layout, str):
        if layout in nx.layout.__all__:
            return eval('nx.{}'.format(layout))
        elif layout == 'coordinates':
            if nf_name is None:
                raise ValueError('nf_name cannot be None')
            return coordinates_layout_closure(nf_name)
        else:
            raise ValueError('layout must be in nx.layout.__all__ or \'coordinates\'')



from __future__ import absolute_import

import numpy as np
from scipy import sparse as sp
from scipy.sparse.linalg import ArpackNoConvergence


def degree(adj):
    """
    Computes the degree matrix of the given adjacency matrix.
    :param adj: rank 2 array or sparse matrix
    :return: the degree matrix in sparse DIA format
    """
    degrees = np.array(adj.sum(1)).flatten()
    if sp.issparse(adj):
        D = sp.diags(degrees)
    else:
        D = np.diag(degrees)
    return D


def degree_power(adj, pow):
    """
    Computes \(D^{p}\) from the given adjacency matrix. Useful for computing
    normalised Laplacians.
    :param adj: rank 2 array or sparse matrix
    :param pow: exponent to which elevate the degree matrix
    :return: the exponentiated degree matrix in sparse DIA format
    """
    degrees = np.power(np.array(adj.sum(1)), pow).flatten()
    degrees[np.isinf(degrees)] = 0.
    if sp.issparse(adj):
        D = sp.diags(degrees)
    else:
        D = np.diag(degrees)
    return D


def normalized_adjacency(adj, symmetric=True):
    """
    Normalizes the given adjacency matrix using the degree matrix as either
    \(D^{-1}A\) or \(D^{-1/2}AD^{-1/2}\) (symmetric normalization).
    :param adj: rank 2 array or sparse matrix;
    :param symmetric: boolean, compute symmetric normalization;
    :return: the normalized adjacency matrix.
    """
    if symmetric:
        normalized_D = degree_power(adj, -0.5)
        output = normalized_D.dot(adj).dot(normalized_D)
    else:
        normalized_D = degree_power(adj, -1.)
        output = normalized_D.dot(adj)
    return output


def laplacian(adj):
    """
    Computes the Laplacian of the given adjacency matrix as \(D - A\).
    :param adj: rank 2 array or sparse matrix;
    :return: the Laplacian.
    """
    return degree(adj) - adj


def normalized_laplacian(adj, symmetric=True):
    """
    Computes a  normalized Laplacian of the given adjacency matrix as
    \(I - D^{-1}A\) or \(I - D^{-1/2}AD^{-1/2}\) (symmetric normalization).
    :param adj: rank 2 array or sparse matrix;
    :param symmetric: boolean, compute symmetric normalization;
    :return: the normalized Laplacian.
    """
    if sp.issparse(adj):
        I = sp.eye(adj.shape[-1], dtype=adj.dtype)
    else:
        I = np.eye(adj.shape[-1], dtype=adj.dtype)
    normalized_adj = normalized_adjacency(adj, symmetric=symmetric)
    return I - normalized_adj


def rescale_laplacian(L, lmax=None):
    """
    Rescales the Laplacian eigenvalues in [-1,1], using lmax as largest eigenvalue.
    """
    if lmax is None:
        try:
            lmax = sp.linalg.eigsh(L, 1, which='LM', return_eigenvectors=False)[0]
        except ArpackNoConvergence:
            lmax = 2
    if sp.issparse(L):
        I = sp.eye(L.shape[-1], dtype=L.dtype)
    else:
        I = np.eye(L.shape[-1], dtype=L.dtype)
    L_scaled = (2. / lmax) * L - I
    return L_scaled


def localpooling_filter(adj, symmetric=True):
    """
    Computes the local pooling filter from the given adjacency matrix, as 
    described by Kipf & Welling (2017).
    :param adj: a np.array or scipy.sparse matrix of rank 2 or 3;
    :param symmetric: boolean, whether to normalize the matrix as
    \(D^{-\\frac{1}{2}}AD^{-\\frac{1}{2}}\) or as \(D^{-1}A\);
    :return: the filter matrix, as dense np.array.
    """
    fltr = adj.copy()
    if sp.issparse(adj):
        I = sp.eye(adj.shape[-1], dtype=adj.dtype)
    else:
        I = np.eye(adj.shape[-1], dtype=adj.dtype)
    if adj.ndim == 3:
        for i in range(adj.shape[0]):
            A_tilde = adj[i] + I
            fltr[i] = normalized_adjacency(A_tilde, symmetric=symmetric)
    else:
        A_tilde = adj + I
        fltr = normalized_adjacency(A_tilde, symmetric=symmetric)
    return fltr


def chebyshev_polynomial(X, k):
    """
    Calculates Chebyshev polynomials up to order k.
    :param X: a np.array or scipy.sparse matrix;
    :param k: the order up to which compute the polynomials,
    :return: a list of k + 1 sparse matrices with one element for each degree of 
            the approximation.
    """
    T_k = list()
    if sp.issparse(X):
        T_k.append(sp.eye(X.shape[0], dtype=X.dtype).tocsr())
    else:
        T_k.append(np.eye(X.shape[0], dtype=X.dtype))
    T_k.append(X)

    def chebyshev_recurrence(T_k_minus_one, T_k_minus_two, X):
        if sp.issparse(X):
            X_ = sp.csr_matrix(X, copy=True)
        else:
            X_ = np.copy(X)
        return 2 * X_.dot(T_k_minus_one) - T_k_minus_two

    for i in range(2, k + 1):
        T_k.append(chebyshev_recurrence(T_k[-1], T_k[-2], X))

    return T_k


def chebyshev_filter(adj, k, symmetric=True):
    """
    Computes the Chebyshev filter from the given adjacency matrix, as described
    in Defferrard et at. (2016).
    :param adj: a np.array or scipy.sparse matrix;
    :param k: integer, the order up to which to compute the Chebyshev polynomials;
    :param symmetric: boolean, whether to normalize the matrix as
    \(D^{-\\frac{1}{2}}AD^{-\\frac{1}{2}}\) or as \(D^{-1}A\);
    :return: a list of k+1 filter matrices, as np.arrays.
    """
    normalized_adj = normalized_adjacency(adj, symmetric)
    if sp.issparse(adj):
        I = sp.eye(adj.shape[0], dtype=adj.dtype)
    else:
        I = np.eye(adj.shape[0], dtype=adj.dtype)
    L = I - normalized_adj  # Compute Laplacian

    # Rescale Laplacian
    L_scaled = rescale_laplacian(L)

    # Compute Chebyshev polynomial approximation
    T_k = chebyshev_polynomial(L_scaled, k)

    return T_k


